require "lxp"

--[[

Parse an XML document into a tree representation.  Uses an
event-driven (SAX-style) XML parser.

In general, call as tree = LxpTree.parse(content)

The format of tree is not particularly Lua-like.  It came from habits
from MOO programming, where there are no tables, only lists.

  <abc/> = {"abc"}
  <abc>d</abc> = {"abc", "d"}
  <abc e='f'>d</abc> = {"abc", "d", attr={e="f"}}
  <abc><h/></abc> = {"abc", {"h"}}

As a special hack for XML-RPC, a mode to ignore character data except
for named elements is available.  Call p=LxpTree.makeParser(), and set
p.filter_chardata true.  Tags not named in the p.keep_chardata table
will have no chardata reported.  You probably want to set
p.merge_chardata as well; this merges all character data in a tag.

]]

local Public = {}
LxpTree = Public

local function top(l)
  return l[table.getn(l)]
end

function Public.makeParser()
  local self = {}
  local callbacks = {}

  function callbacks.StartElement(parser, elementName, attributes)
    local t = {elementName}
    if attributes and attributes[1] then t.attr = attributes end
    self.str = nil
    table.insert(top(self.stack), t)
    table.insert(self.stack, t)
    if self.filter_chardata and self.keep_chardata[elementName] then
      self.want_chardata = true
    end
  end

  function callbacks.EndElement(parser, elementName)
    if self.merge_chardata and self.str and self.keep_chardata[elementName] then
      table.insert(top(self.stack), self.str)
    end
    self.str = nil
    self.want_chardata = false
    table.remove(self.stack)
  end

  function callbacks.CharacterData(parser, string)
    if self.filter_chardata and not self.want_chardata then return end
    if self.merge_chardata then
      if self.str then
        self.str = self.str .. string
      else
        self.str = string
      end
    else
      table.insert(top(self.stack), string)
    end
  end
  
  local parser = lxp.new(callbacks)
  self.stack = {{}}
  
  function self:parse(s)
    local result, msg, line, col, pos = parser:parse(s)
    if result then
      result, msg, line, col, pos = parser:parse()
    end
    if not result then
      error("expat parse error "..msg.." at line "..line.." column "..col)
    end
    parser:close()
    return self.stack[1][1]
  end

  return self
end

function Public.parse(s)
  local p = Public.makeParser()
  return p:parse(s)
end

real_tostring = real_tostring or tostring
function tostring(x)
  local s
  if type(x) == "table" then
    s = "{"
    local i, v = next(x)
    while i do
      s = s .. tostring(i) .. "=" .. tostring(v)
      i, v = next(x, i)
      if i then s = s .. "," end
    end
    return s .. "}"
  else return real_tostring(x)
  end
end
