--[[

Given a tree representation of an XML document, serialize to a string.
See lxptree for documentation on the structure of trees expected.

This uses the string buffering strategy from LTN 9 to avoid O(n^2)
behavior for string appends.
]]

local Public = {}
XMLGen = Public

-- Right out of LTN 9

local function newBuffer ()
  return {n=0}     -- 'n' counts number of elements in the stack
end

local function addString (stack, s)
  table.insert(stack, s)       -- push 's' into the top of the stack
  for i=stack.n-1, 1, -1 do
    if string.len(stack[i]) > string.len(stack[i+1]) then break end
    stack[i] = stack[i]..table.remove(stack)
  end
end

local function toString (stack)
  for i=stack.n-1, 1, -1 do
    stack[i] = stack[i]..table.remove(stack)
  end
  return stack[1]
end

-- It might be clever to do this in a single pass
function Public.quote(s)
  s = string.gsub(s, "&", "&amp;")
  s = string.gsub(s, "<", "&lt;")
  s = string.gsub(s, ">", "&gt;")
  return s
end

function Public.quoteForSingle(s)
  s = Public.quote(s)
  s = string.gsub(s, "'", "&apos;")
  return s
end

local function marshallToBuffer(s, l)
  local name = l[1]
  if l.attr then
    addString(s, "<"..name)
    for i,v in l.attr do
      addString(s, " "..i.."='"..Public.quoteForSingle(v).."'")
    end
    addString(s, ">")
  else
    addString(s, "<"..name..">")
  end
  for i = 2,table.getn(l) do
    local elt = l[i]
    if type(elt) == "table" then
      marshallToBuffer(s, elt)
    elseif type(elt) == "string" then
      addString(s, Public.quote(elt))
    else
      error("non-table, non-string in marshalling")
    end
  end
  addString(s, "</"..name..">")
end

function Public.treeToString(t)
  local s = newBuffer()
  marshallToBuffer(s, t)
  return toString(s)
end

