require "xmlgen"
require "luasocket"
require "lxptree2"

--[[

XMLRPC: an XML-RPC binding for Lua.

See http://www.xmlrpc.org/ for information on XML-RPC.

XML-RPC's types don't map directly to the Lua types.  A table with the
key XMLRPC.type is a special table.  The value of the key describes
which XML-RPC type the table represents.  To construct these special
values, you should use their constructor functions rather than table
constructors.

<string>: Lua string
<i4>, <integer>: Lua number
<array>: a Lua table with a key of 1 present, or with ["n"]=0
<struct>: a Lua table (without the type key)

<boolean>: true or false

<double>: Build with XMLRPC.double(123.78); access with v.double
<base64>: Build with XMLRPC.base64("BaSe64--"); access with v.base64
<dateTime.iso8601>: Build with XMLRPC.dateTime(string); access with v.dateTime

Aside from that, the examples are the documentation right now.  Have fun.

Jay Carlson
nop@nop.com
]]

local Public = {}
XMLRPC = Public

-- An abortive attempt at getting rid of the Private table.  In 5.0 you
-- can just replace Private with local...except that you have to forward-
-- declare everything.  Oh well.  I don't recommend this strategy.

local build_value, build_array, build_struct, build_special
local parse_value

-- A unique table to designate other special types
Public.type = {"XMLRPC_Type"}

Public.USERAGENT = "Lua XMLRPC/0.5"

Public.strict = true

Public.registry = {}

local keep_chardata = {
  i4=true, int=true, boolean=true, string=true, double=true, 
  ["dateTime.iso8601"]=true, base64=true, methodName=true,
  name=true, value=true
}

local getn = table.getn
local tinsert = table.insert
local tremove = table.remove
local strfind = string.find

local function findChild(l, name)
  for i=2,getn(l) do
    if l[i][1] == name then
      return l[i]
    end
  end
  return nil
end

local function expect(l, name)
  if (l[1] ~= name) then
    error("expected " .. name .. " but found " .. l[1])
  end
end

local function parse_params(l)
  expect(l, "params")
  local params = {}
  for i = 2,getn(l) do
    local param_t = l[i]
    expect(param_t, "param")
    tinsert(params, parse_value(param_t[2]))
  end
  return params
end

function Public.parse_methodCall_tree(l)
  local mc = {}
  expect(l, "methodCall")
  local mn = findChild(l, "methodName")
  expect(mn, "methodName")
  mc.name = mn[2]
  local params_tree = findChild(l, "params")
  if (params_tree) then
    mc.params = parse_params(params_tree)
  end
  return mc
end


local parse_value_type = {}

parse_value = function (l) -- forward-declared
  expect(l, "value")
  -- if not debug1 then write(_STDERR, "----"..tostring( l).."\n") flush() end
  if type(l[2]) == "string" then return l[2] end
  local type = l[2][1]
  f = parse_value_type[type]
  if not f then error("unimplemented value type " .. type) end
  return f(l[2])
end

function parse_value_type.i4(l)
  expect(l, "i4")
  return tonumber(l[2])
end

function parse_value_type.int(l)
  expect(l, "int")
  return tonumber(l[2])
end

function parse_value_type.boolean(l)
  expect(l, "boolean")
  if tonumber(l[2]) == 0 then
    return false
  elseif tonumber(l[2]) == 1 then
    return true
  else
    error("invalid boolean value " .. l[2])
  end
end

function parse_value_type.string(l)
  expect(l, "string")
  return l[2] or ""
end

function Public.double(n)
  return {[Public.type]="double", double=n}
end

function parse_value_type.double(l)
  expect(l, "double")
  if Public.strict then
    return Public.double(tonumber(l[2]))
  else
    return tonumber(l[2])
  end
end

function Public.dateTime(s)
  return {[Public.type]="dateTime", dateTime=s}
end

parse_value_type["dateTime.iso8601"] = 
  function (l)
    expect(l, "dateTime.iso8601")
    if Public.strict then
      return Public.dateTime(l[2])
    else
      return l[2]
    end
  end

function Public.base64(s)
  return {[Public.type]="base64", base64=s}
end

function parse_value_type.base64(l)
  expect(l, "base64")
  return Public.base64(l[2])
end

function parse_value_type.array(l)
  expect(l, "array")
  local data = l[2]
  expect(data, "data")
  local elts = {n=0}
  for i = 2,getn(data) do
    tinsert(elts, parse_value(data[i]))
  end
  return elts
end

local function parse_member(l)
  expect(l, "member")
  local name_tree = findChild(l, "name")
  local value_tree = findChild(l, "value")
  assert(name_tree and value_tree, "<member> must contain <name> and <value> nodes")

  local name = name_tree[2]
  local value = parse_value(value_tree)
  return name,value
end

function parse_value_type.struct(l)
  expect(l, "struct")
  local struct = {}
  for i = 2,getn(l) do
    local name,value = parse_member(l[i])
    struct[name] = value
  end
  return struct
end

function Public.parse_methodResponse_tree(l)
  expect(l, "methodResponse")
  if (l[2][1] == "params") then
    return parse_params(l[2])
  elseif (l[2][1] == "fault") then
    local val = parse_value(l[2][2])
    return nil,val
  end
end

local function buildtree(methodname, params)
  local base = {"methodCall", {"methodName", methodname}}
  local param_tree = {"params"}
  for i=1,getn(params) do
    local p = params[i]
    tinsert(param_tree, {"param", {"value", build_value(p)}})
  end
  if (getn(params) ~= 0) then
    tinsert(base, param_tree)
  end
  return base
end

local function buildresponsetree(params)
  local base = {"methodResponse"}
  local param_tree = {"params"}
  for i=1,getn(params) do
    local p = params[i]
    tinsert(param_tree, {"param", {"value", build_value(p)}})
  end
  tinsert(base, param_tree)
  return base
end

local function buildfaulttree(fault_struct)
  local base = {"methodResponse"}
  tinsert(base, {"fault", {"value", build_value(fault_struct)}})
  return base
end

build_value = function (p)
  if p == true then
    return {"boolean", "1"}
  elseif p == false then
    return {"boolean", "0"}
  elseif type(p) == "number" then
    return {"i4", tostring(p)}
  elseif type(p) == "string" then
    return {"string", p}
  elseif type(p) == "table" then
    if p[1] or p.n == 0 then
      -- treat as array
      return build_array(p)
    elseif p[Public.type] then
      return build_special(p)
    else
      return build_struct(p)
    end
  elseif p == nil then
    error("marshalling nil (as false) is no longer supported")
  else
    error("unsupported transfer format")
  end
end

build_array = function(p)
  local values = {"data"}
  for i=1,getn(p) do
    local elt = p[i]
    local elt_tree = {"value", build_value(elt)}
    tinsert(values, elt_tree)
  end
  return {"array", values}
end

build_struct = function(p)
  local tree = {"struct"}
  for i,v in p do
    local elt_tree = {"member", {"name", i}, {"value", build_value(v)}}
    tinsert(tree, elt_tree)
  end
  return tree
end

build_special = function (p)
  local t = p[Public.type]
  if t == "double" then
    return {"double", tostring(p.double)}
  elseif t == "dateTime" then
    return {"dateTime.iso8601", p.dateTime}
  elseif t == "base64" then
    return {"base64", p.base64}
  else
    error("unknown special type")
  end
end

function Public.callToString(methodname, params)
  local tree = buildtree(methodname, params)
  return XMLGen.treeToString(tree)
end

function Public.responseToString(params)
  -- print(build_value({1}))
  local tree = buildresponsetree(params)
  return XMLGen.treeToString(tree)
end

function Public.faultToString(fault_struct)
  local tree = buildfaulttree(fault_struct)
  return XMLGen.treeToString(tree)
end

----- now, to actually parse stuff

local function makeParser()
  local p = LxpTree.makeParser()
  p.filter_chardata = true
  p.merge_chardata = true
  p.keep_chardata = keep_chardata
  return p
end

function Public.parseMethodCall(s)
  local p = makeParser()
  local tree = p:parse(s)
  return Public.parse_methodCall_tree(tree)
end

function Public.parseMethodResponse(s)
  local p = makeParser()
  local tree = p:parse(s)
  return Public.parse_methodResponse_tree(tree)
end


-- Now, for really silly syntactic sugar

-- some unique tables
local url_key = {"url_key"}
local name_key = {"name_key"}

local server_metatable = {}
local func_metatable = {}

function server_metatable.__index(table, key)
  local root_url = table[url_key]
  local f = {[url_key]=root_url, [name_key]=key}
  setmetatable(f, func_metatable)
  return f
end

function func_metatable.__index(table, key)
  local root_url = table[url_key]
  local basename = table[name_key]
  local f = {[url_key]=root_url, [name_key]=basename.."."..key}
  setmetatable(f, func_metatable)
  return f
end

function Public.Server(url)
  local s = {[url_key]=url}
  setmetatable(s, server_metatable)
  return s
end

function Public.do_xml_rpc_call(url, host, s)
   local body, headers, status, msg = socket.http.post{
      url=url, 
      headers={["content-type"]="text/xml", ["user-agent"]=Public.USERAGENT},  
      body=s}
    -- print("body:", body)
    -- print("headers:", headers)
    -- print("status: ", status)
    -- print("error: ", msg)
    -- print("-------------")
   
   if msg or status ~= 200 then
     error("XMLRPC http error "..(status or msg or "NO-STATUS")..", body follows:\n"..(body or "(null body)"))
   end

   -- print(body)
   
   local res, err = XMLRPC.parseMethodResponse(body)
   if err then
      error("XMLRPC error code: "..err.faultCode.." with text "..err.faultString)
   else
      return unpack(res)
    end
end


function func_metatable.__call(func, ...)
  local url = func[url_key]
  -- print("url", rawget(func, "url"))
  -- print("name", rawget(func, "name"))
  local s = "<?xml version='1.0'?>"..Public.callToString(func[name_key], arg)
  local _,_,host = strfind(url, "^[a-zA-Z]+:[0-9]*//([^/]+)")

  return Public.do_xml_rpc_call(url, host, s)
end

local function pack(...)
   return arg
end

function findFunction(name)
  local root = Public.registry
  while true do
    local match,_,first,second = strfind(name, "([^.]+)%.(.+)")
    if not match then
      return root and root[name] 
    else
      if not root[first] then return nil end
      root = root[first]
      name = second
    end
  end
end

local function internalDispatchString(body)
   local mc_info = Public.parseMethodCall(body)
   local func = findFunction(mc_info.name)

   if not func then error("no such function "..mc_info.name) end

   local params = mc_info.params or {}
   local result = pack(func(unpack(params)))
   local result_string = XMLRPC.responseToString(result).."\n"
   return result_string
end

local function pack_rest(v, ...)
  return v, args
end

function Public.dispatchString(body)
   local bomb_message = {}
   -- file = file or _OUTPUT
   local succeeded, result_string = pcall(internalDispatchString, body)
   if not succeeded then
     result_string = XMLRPC.faultToString({faultCode=1, faultString=result_string}).."\n"
   end
   return result_string
end


